/**
 * File:   cObject.h
 * Project: LinWarrior 3D
 * 
 * Created on March 28, 2008, 10:23 PM (1999)
 *
 * Generic Game-Object with as much common
 * attributes as making sense and as needed
 * for various World-Filtering (Worldsense for Objects).
 */

#ifndef _OBJECT_H
#define _OBJECT_H

class cObject;
class cPad;
class cController;

#include "cWorld.h"
#include <string>

/**
 * Fundamental function for collision-check.
 */
float collideLineWithAABB(
        float* start3fv,
        float* end3fv,
        float* min3fv,
        float* max3fv,
        float* impactpoint3fv,
        float* impactnormal3fv);

/**
 * Generic Game-Object (Object with Roles) which
 * collects as much common attributes as are needed
 * by the different roles and as needed by World-Filtering
 * (Worldsense for subclasses implementing the roles).
 * This object does again hold as much methods as needed
 * but without (real) implementation (which is subclass-responsibility).
 * Nevertheless every method has a empty body or a body returning
 * some default value, so that callers need not care about
 * the kind of the object (roles) too much.
 */
class cObject {
public:
    friend class cWorld;

    static int ENABLE_TEXTURE_3D;

    /**
     * Enumeration of entity body part units: Body, Legs, Left, Right.
     */
    enum Parts {
        BODY = 0, // Some Objects only have this
        LEGS,
        LEFT,
        RIGHT,
        MAX_PARTS
    };

    /**
     * Enumeration of roles an object may assume.
     */
    enum Roles {
        // Roles indicating Attributes
        NAMEABLE,
        SOCIALISED,
        TRACEABLE,
        COLLIDEABLE,
        DAMAGEABLE,
        ENTITY,

        // Roles indicating Kind of Objects
        MECH,
        BUILDING,
        TREE,
        IMMOBILE,

        // Roles indicating belonging to certain Social-Parties
        RED,
        BLUE,
        GREEN,
        YELLOW,
        CIVIL,

        // Roles indicating State
        WOUNDED, //  <= 75% hp
        SERIOUS, // <= 50% hp
        CRITICAL, // <= 25% hp
        DEAD, // <= 0% hp, note that the later include the former.

        // Special Roles
        HUMANPLAYER,
        FRAGGED,
    };

public:
    // BASE

    /// Unique Object ID, simulation date:time:serial of spawn.
    tSerial mSerial;
    /// Bitmask, lists the roles this object plays.
    unsigned long mRoles;
    /// Simulated Seconds since spawn, update in animateSelf!
    double mSeconds;

    // NAMEABLE

    /// Name of object.
    std::string mName;
    /// String describing the object.
    std::string mDescription;
    /// { 0, .., 25 } for { a, ..., z }.
    unsigned int mDesignation;

    // SOCIALISED

    /// future: Bitmask, lists Roles which Allies have.
    unsigned long mAllies;
    /// Bitmask, lists Roles which Enemies play.
    unsigned long mEnemies;

    // TRACEABLE

    /// Acceleration Vector.
    float mAcc[3];
    /// Velocity Vector.
    float mVel[3];
    /// Position Vector.
    float mPos[3];
    /// Rotation around three axes.
    float mRot[3];
    /// Size as radius.
    float mRadius;

    // DAMAGEABLE

    /// Health-Points for Body-Parts.
    float mHp[MAX_PARTS];
    // Splasharmor,Beamarmor,Piercearmor,Splasharmor,Beamarmor, Piercearmor
    // Heat,Heatsinks

    // ENTITY
    
    /// Aim target
    tSerial mTarget;
    /// Movement target
    float mDestination[3];
    /// Human, AI/Remote Pad-Control over the Object.
    cPad* mPad;
    /// Command Stack controlls the pad when enabled.
    cController* mController;

public:
    cObject();
    cObject(cObject* source);

    virtual ~cObject() {
    }

    // Roles

    /// Check wether this Object is playing the given role.
    bool hasRole(Roles role) {
        return (mRoles & (1UL << role));
    }

    /// Add a role this objects has to play.
    void addRole(Roles role) {
        mRoles |= (1UL << role);
    }

    /// Remove a role this object will not play anymore.
    void remRole(Roles role) {
        mRoles &= ~(1UL << role);
    }

    // Allies

    /// Check wether an object playing a certain role is an ally.
    bool hasAllied(Roles role) {
        return (mAllies & (1UL << role));
    }

    /// Add roles which allies of this object play.
    void addAllied(Roles role) {
        mAllies |= (1UL << role);
    }

    /// Remove objects playing a certain role from ally list.
    void remAllied(Roles role) {
        mAllies &= ~(1UL << role);
    }

    // Enemies

    /// Check wether an object playing a certain role is an enemy.
    bool hasEnemy(Roles role) {
        return (mEnemies & (1UL << role));
    }

    /// Add roles which enemies of this object play.
    void addEnemy(Roles role) {
        mEnemies |= (1UL << role);
    }

    /// Remove objects playing a certain role from enemy list.
    void remEnemy(Roles role) {
        mEnemies &= ~(1UL << role);
    }

    /// Called right after object was spawned into the world.
    virtual void onSpawn() {
    }

    /// Called right after object has been removed from object list and index.
    virtual void onFrag() {
    }

    /// glMultiplies in the Object's camera matrix (and set Listener position).
    virtual void multEyeMatrix() {
    }

    /// Advance internal timers,animation state and pose, check gamepad.
    virtual void animate(float spf) {
        mSeconds += spf;
    }

    /// Precalculate neccessary transformations - matrices, mountpoints, pos ..
    virtual void transform() {
    }

    /// Render solid non-translucent parts of the object.
    virtual void drawSolid() {
    }

    /// Render translucent object-parts and visual effects.
    virtual void drawEffect() {
    }

    /// Render HUD-contents as seen when looking through the objects-eyes.
    virtual void drawHUD() {
    }

    // Receiving damage:

    /**
     * Modells an impact with this object.
     */
    struct rImpact {
        float pos[3];
        float nrm[3];
    };

    /**
     * Deal damage to the object (parts) (if applicable) when possible at
     * intersection with the described cylinder/beam, fallback to a line
     * or other approximation as you wish.
     * all: deal damage to anything intersected not just the nearest part.
     */
    virtual rImpact* damageByBeam(float* a, float* b, float radius, float damage, bool all) {
        return NULL;
    }

    // Deal damage to object (parts) inside the described sphere.
    // all: deal damage to anything intersected not just the nearest part.
    //virtual rImpact* damageBySphere(float* a, float radius, float damage, bool all) { return NULL; };

    // Deal damage to object (parts) without respect to any location.
    //virtual void damage(float damage) {};


    // Physics for Traceable:

    /// Returns distance to nearest objects straight below or own height.
    virtual float checkRelHeight();

    /**
     * Check for blocking objects in this objects velocity-path within the
     * next seconds and adjust this object's velocity to prevent intersection
     * and come to a halt at the collision point.
     */
    virtual float checkAhead(float seconds);


    // Related to States:

    // Queries

    /// Search through world-database to find nearest enemy, returns OID.
    virtual tSerial enemyNearby() {
        return 0;
    }

    // Fuzzy Predicates [0,1]

    /// Has this object reached it's destination position? Fuzzyvalue in [0,1].
    virtual float inDestinationRange() {
        return 0.0f;
    }

    /// Is this object in "hand-to-hand"-combat range? Fuzzyvalue in [0,1].
    virtual float inMeeleRange() {
        return 0.0f;
    }

    /// Is this object in firing range. Fuzzyvalue in [0,1].
    virtual float inWeaponRange() {
        return 0.0f;
    }

    /// Is this object in targeting/visual distance? Fuzzyvalue in [0,1].
    virtual float inTargetRange() {
        return 0.0f;
    }
    
    /*
    virtual float blockedAhead() {
        return 0.0f;
    }
    
    virtual float blockedLeft() {
        return 0.0f;
    }

    virtual float blockedRight() {
        return 0.0f;
    }

     */

    // Output Actions for Entities

    /// Move towards move-target as fast as possible.
    virtual void do_moveTowards() {
    }
    
    /// Move to move-target throttling down as target comes near.
    virtual void do_moveNear() {
    }

    /// Aim at aim-target.
    virtual void do_aimAt() {
    }

    /// Fire at aim-target - ie. just fire where object is aiming at right now.
    virtual void do_fireAt() {
    }

    /// Nothing to do for object - possibly idle animation.
    virtual void do_idle() {
    }

    /// Set aim towards given object - ie. set aim-target.
    virtual void do_aimFor(tSerial target) {
    }

    /// Set aim towards given location. Fixme: There should be a differentiation between walk- and aim-target.
    virtual void do_aimFor(float* dest) {
    }
};

#endif
